package ch.hepia.numeric;
import java.util.List;
import java.util.ArrayList;
import java.util.function.DoubleFunction;
import java.util.function.Function;

public class Transposed{
    private List<Double> vecT;

    public Transposed(){
        this.vecT = new ArrayList<>();
    };

    private Transposed(Double... elements) {
        List<Double> vec = new ArrayList<>();
        for (double elem : elements){vec.add(elem);}
        this.vecT = vec;
    }

    public Transposed(List<Double> elements){
        this.vecT = elements;
    }

    public Transposed(Vector That){
        this.vecT = new ArrayList<>();
        for (double elem : That.getValues()){vecT.add(elem);}
    };

     public List<Double> getValues(){
        return this.vecT;
    }
    
    void set(int i, double value) {
        this.vecT.set(i, value);
    }

    public int len() {
        return this.getValues().size(); 
    }

    public double get(int i) {
        return this.getValues().get(i);
    }
    
    public double dot(Vector That){
        double res = 0.0;
        Integer count = 0;
        for (double coo : this.vecT) {
            res = res + (coo * That.get(count)); 
            count = count + 1;
        }
        return res;
        //throw new UnsupportedOperationException("This feature isn't implemented yet");
    }
    
    public Transposed add(Transposed that) {
        List<Double> newVec = new ArrayList<>();
        for(int i=0; i < this.len();i++){
            newVec.add(this.get(i)+that.get(i));    
        }
        return new Transposed(newVec);
    }

    public static Transposed zeros(int nb) {
        List<Double> newVec = new ArrayList<>();
        for (int i = 0; i < nb ; i++){
            newVec.add(0.0);
        }
        return new Transposed(newVec);
    }

    public static Transposed sum(List<Transposed> vs) {
        Transposed totalVec = Transposed.zeros(vs.get(0).len());
        for (Transposed vec : vs){
            totalVec = totalVec.add(vec);
        }
        return totalVec;
    }

    public Transposed sub(Transposed That){
        Transposed resT = new Transposed();
        Integer count = 0;
        for (double coo : this.getValues()) {
            resT.vecT.add(coo - That.getValues().get(count));
            count = count + 1;
        }
        return resT;
        //throw new UnsupportedOperationException("This feature isn't implemented yet");
    };

    public Transposed mul(double m) {
        List<Double> newVec = new ArrayList<>();
        for (int i=0; i<this.len();i++){
            newVec.add(this.get(i)*m);
        }
        return new Transposed(newVec);
    }

    public static Transposed of(Double... elements) {
        return new Transposed(elements);
    }

    public static Transposed of(List<Double> elements) {
        return new Transposed(elements);
    }

    public static Transposed empty() {
        return new Transposed();
    }

    public double norm() {
        double sum =  0.0;     
        for (int i=0; i<this.len();i++){
            sum += Math.pow(this.get(i),2.0);
        }
        return Math.sqrt(sum);
    }

    public static double norms(List<Transposed> vs) {
        return Transposed.sum(vs).norm();
    }

    public Transposed map(DoubleFunction<Double> f) {
        List<Double> accVec = this.getValues();
        List<Double> newVec = new ArrayList<>();
    
        for (Double val : accVec){
            newVec.add(f.apply(val));
        }
        return new Transposed(newVec);
    }

    public static Transposed ones(int nb) {
        List<Double> newVec = new ArrayList<>();
        for (int i = 0; i < nb ; i++){
            newVec.add(1.0);
        }
        return new Transposed(newVec);
    }

    public static Transposed fill(int nb, double value) {
        List<Double> newVec = new ArrayList<>();
        for (int i = 0; i < nb ; i++){
            newVec.add(value);
        }
        return new Transposed(newVec);
    } 

    public static Transposed tabulate(int nb, Function<Integer, Double> f) {
        List<Double> newVec = new ArrayList<>();
        for (int i=0; i < nb; i++){
            newVec.add(f.apply(i));
        }
        return new Transposed(newVec);
    }

    public Transposed concat(Transposed that) {
        List<Double> newVec = this.getValues();
        newVec.addAll(that.getValues());
        return new Transposed(newVec);
    }

    public Transposed sliceFrom(int i) {
        List<Double> newVec = new ArrayList<>();
        for (int j=0; j < this.len();j++){
            if (j >= i){
                newVec.add(this.get(j));
            }
        }
        return new Transposed(newVec);
    }

    public Transposed sliceTo(int i) {
        List<Double> newVec = new ArrayList<>();
        for (int j=0; j < this.len();j++){
            if (j < i){
                newVec.add(this.get(j));
            }
        }
        return new Transposed(newVec);
    }

    public Transposed slice(int from, int to) {
        Transposed newVec = this;
        newVec = newVec.sliceFrom(from);
        newVec = newVec.sliceTo(to-from);
        return newVec;
    }

    public Transposed removed(int i) {
        List<Double> newVec = this.getValues();
        newVec.remove(i);
        return new Transposed(newVec);
    }
    @Override
    public String toString() {
        String tr = "Transposed[";
        for (double val: this.getValues()){
            tr = tr.concat(val + ", ");
        }
        tr = tr.substring(0, tr.length()-2).concat("]");
        return tr;
    }

    @Override
    public boolean equals(Object obj){
        if (this == obj) return true;
        if (obj == null || obj.getClass() != this.getClass()) {
            return false;
        }
        Transposed tra = (Transposed) obj;
        if (tra.len() != this.len()){
            return false;
        }
        for (int i=0;i<tra.len();i++){
            if (Double.compare(tra.getValues().get(i), this.getValues().get(i))!=0){
                return false;
            }
        }
        return true;
    }





}